# -*- coding: utf-8 -*-

# pylint: disable=invalid-name

import asyncio
import fcntl
import json
import os
import struct
import termios

from django.http import HttpResponse
from django.views import generic

from django_sivumedia import JSBool
from pistoke.nakyma import WebsocketNakyma

from .paate import Paateprosessi
from .vimpain import (
  XtermJS,
  XtermAddonFit,
  XtermAddonSearch,
  XtermAddonWebLinks,
)


class XtermNakyma(
  XtermAddonFit,
  XtermAddonSearch,
  XtermAddonWebLinks,
  XtermJS,
  WebsocketNakyma,
  generic.TemplateView
):
  '''
  Django-näkymä vuorovaikutteisen Websocket-yhteyden tarjoamiseen.

  Käyttöliittymän luontiin käytetään Xterm.JS-vimpainta.
  '''
  template_name = 'xterm/xterm.html'

  class Media:
    js = ['xterm/js/xterm.js']

  # Xterm-ikkunan alustusparametrit.
  xterm = {
    'cursorBlink': JSBool(True),
    'macOptionIsMeta': JSBool(False),
    'scrollback': 5000,
  }

  def prosessi(self):
    ''' Päätteessä ajettava prosessi. '''
    raise NotImplementedError

  def get(self, request, *args, **kwargs):
    '''
    Varmista, että Websocket-protokolla on saatavilla.
    '''
    if getattr(request, 'websocket', None) is None:
      return HttpResponse(status=503)
    return super().get(request, *args, **kwargs)
    # def get

  @staticmethod
  async def _lahetys(send, jono):
    '''
    Asynkroninen lähetys: odota dataa jonosta ja lähetä.
    '''
    while True:
      data = await jono.get()
      await send(data.decode())
      jono.task_done()
      # while True
    # async def _lahetys

  @staticmethod
  async def _vastaanotto(receive, fd):
    '''
    Asynkroninen lukutehtävä: lue dataa ja syötä prosessille.
    '''
    while True:
      data = await receive()

      if isinstance(data, bytes):
        # Binäärisanoma: näppäinkoodi.
        # Control-C katkaisee syötteen.
        if data == b'\x03':
          break
        try:
          os.write(fd, data)
        except OSError:
          break

      elif isinstance(data, str):
        # Tekstisanoma: JSON-muotoinen IOCTL-ohjauskomento.
        data = json.loads(data, strict=False)
        if 'cols' in data and 'rows' in data:
          # Asetetaan ikkunan koko.
          fcntl.ioctl(
            fd,
            termios.TIOCSWINSZ,
            struct.pack("HHHH", data['rows'], data['cols'], 0, 0)
          )
      # while True
    # async def _vastaanotto

  async def websocket(self, request, *args, **kwargs):
    # pylint: disable=unused-argument
    lahetettava_data = asyncio.Queue()

    def tulosteen_vastaanotto(prosessi):
      '''
      Callback-tyyppinen rutiini datan lukemiseksi PTY:ltä.
      Kerää kaikki saatavilla oleva data, lisää lähetysjonoon.
      '''
      data = bytearray()
      while True:
        try:
          data += os.read(prosessi.fd, 4096)
        except (IOError, BlockingIOError):
          break
        if not data:
          return
      lahetettava_data.put_nowait(data)
      # def tulosteen_vastaanotto

    # Aloita tulostedatan lähetys itsenäisenä tehtävänä.
    lahetys_tehtava = asyncio.ensure_future(
      self._lahetys(request.send, lahetettava_data)
    )

    # Alusta pääteprosessi ja datan vastaanottotehtävä.
    paate = await Paateprosessi(self.prosessi, lukija=tulosteen_vastaanotto)
    tehtavat = {
      paate,
      asyncio.ensure_future(
        self._vastaanotto(request.receive, paate.fd)
      ),
    }

    # Odota siksi kunnes joko syöte katkaistaan
    # tai pääteprosessi on valmis.
    try:
      _, tehtavat = await asyncio.wait(
        tehtavat, return_when=asyncio.FIRST_COMPLETED
      )

    # Peruuta kesken jääneet tehtävät ja odota ne loppuun.
    finally:
      for kesken in tehtavat:
        kesken.cancel()
      await asyncio.gather(*tehtavat, return_exceptions=True)
      # Odota siksi, kunnes kaikki data on lähetetty.
      await lahetettava_data.join()
      # Peruuta lähetystehtävä ja odota se loppuun.
      lahetys_tehtava.cancel()
      await asyncio.gather(lahetys_tehtava, return_exceptions=True)

    # async def websocket

  # class XtermNakyma



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name

import asyncio
import os
import pty
import signal
import termios


class Paateprosessi(asyncio.Future):
  ''' Asynkroninen kääre `pty.fork`-prosessin ajamiseen. '''

  _kaynnissa_olevat_prosessit = {}
  _signaalinkasittelija_asennettu = False

  @classmethod
  def _prosessi_paattyi(cls):
    '''
    Käsittelymetodi `SIGCHLD`-signaalille.
    Merkitään ne prosessit päättyneiksi,
    joiden edustama PID on päättynyt.
    '''
    while True:
      try:
        pid, _ = os.waitpid(-1, os.WNOHANG)
      except OSError:
        break
      if pid <= 0:
        break
      prosessi = cls._kaynnissa_olevat_prosessit.pop(pid, None)
      if prosessi is not None and not prosessi.done():
        prosessi.set_result(0)
      # while True
    # def _prosessi_paattyi

  @classmethod
  def _asenna_signaalinkasittelija(cls):
    '''
    Asenna signaalinkäsittelijä.
    Huom. ajettava prosessin pääsäikeestä.
    '''
    if not cls._signaalinkasittelija_asennettu:
      oletus_sigchld = signal.getsignal(signal.SIGCHLD)
      def tarkistus(*args):
        cls._prosessi_paattyi()
        if callable(oletus_sigchld):
          oletus_sigchld(*args)
      signal.signal(signal.SIGCHLD, tarkistus)
      cls._signaalinkasittelija_asennettu = True
      # if not cls._signaalinkasittelija_asennettu
    # def _asenna_signaalinkasittelija

  @staticmethod
  # pylint: disable=inconsistent-return-statements
  def _kaynnista_prosessi(prosessi):
    '''
    Avaa uusi prosessi ja suorita annettu funktio.
    Huom. ajettava prosessin pääsäikeestä.
    '''
    pid, fd = pty.fork()
    if pid:
      return pid, fd
    try:
      prosessi()
    finally:
      # Suljetaan prosessi riippumatta mahdollisesta poikkeuksesta.
      os._exit(0) # pylint: disable=protected-access
    # def _kaynnista_prosessi

  async def __new__(cls, *args, **kwargs):
    ''' Asenna signaalinkäsittelijä. '''
    # Asenna globaali signaalinkäsittelijä ensimmäisen prosessin
    # luonnin yhteydessä.
    loop = asyncio.get_event_loop()
    loop.call_soon_threadsafe(cls._asenna_signaalinkasittelija)

    # Luo uusi prosessi asynkronisesti.
    prosessi = super().__new__(cls, *args, **kwargs)
    await prosessi.__init__(*args, **kwargs)
    return prosessi
    # def __new__

  async def __init__(
    self, prosessi, *, lukija=None, kirjoittaja=None, **kwargs
  ):
    '''
    Käynnistä aliprosessi.
    '''
    super().__init__(**kwargs)
    loop = self._loop

    # Irrota erillinen prosessi synkronisesti
    # ajossa olevan ohjelman pääsäikeessä.
    self.pid, self.fd = await loop.run_in_executor(
      None, self._kaynnista_prosessi, prosessi
    )

    # Aseta PTY-master-kahva `non-blocking`-tilaan.
    os.set_blocking(self.fd, False)

    # Lisää lukija ja kirjoittaja PTY-master-kahvalle.
    if lukija is not None:
      loop.add_reader(self.fd, lukija, self)
    if kirjoittaja is not None:
      loop.add_writer(self.fd, kirjoittaja, self)

    # Tallenna prosessi luokkakohtaiselle avointen prosessien listalle.
    self._kaynnissa_olevat_prosessit[self.pid] = self

    # Estetään Control-C-painallusta nostamasta SIGINT-signaalia,
    # joka sammuttaa koko ASGI-prosessin.
    t = termios.tcgetattr(self.fd)
    t[6][termios.VINTR] = 0
    termios.tcsetattr(self.fd, termios.TCSANOW, t)

    # pylint: disable=unused-variable
    @self.add_done_callback
    def done_callback(_):
      # Katkaistaan luku ja kirjoitus PTY:ltä, suljetaan kahva.
      self._loop.remove_reader(self.fd)
      self._loop.remove_writer(self.fd)
      os.close(self.fd)
      # def done_callback

    # def __init__

  def cancel(self):
    if super().cancel():
      try:
        os.kill(self.pid, signal.SIGHUP)
        os.kill(self.pid, signal.SIGKILL)
      except OSError:
        pass
      return True
    return False
    # def cancel

  # class Paateprosessi



from django_sivumedia import Mediasaate


class XtermJS(Mediasaate):
  class Media:
    versio = '5.3.0'
    css = {'all': [
      f'https://cdn.jsdelivr.net/npm/xterm@{versio}/css/xterm.min.css',
    ]}
    js = [
      f'https://cdn.jsdelivr.net/npm/xterm@{versio}/lib/xterm.min.js'
    ]


class XtermAddonFit(XtermJS):
  class Media:
    versio = '0.10.0'
    js = [
      f'https://cdn.jsdelivr.net/npm/@xterm/addon-fit@{versio}'
      f'/lib/addon-fit.min.js'
    ]


class XtermAddonSearch(XtermJS):
  class Media:
    versio = '0.15.0'
    js = [
      f'https://cdn.jsdelivr.net/npm/@xterm/addon-search@{versio}'
      f'/lib/addon-search.min.js'
    ]


class XtermAddonWebLinks(XtermJS):
  class Media:
    versio = '0.11.0'
    js = [
      f'https://cdn.jsdelivr.net/npm/@xterm/addon-web-links@{versio}'
      f'/lib/addon-web-links.min.js'
    ]
